#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat_ws main application module

This module can be used by GUnicorn or Uvicorn to create application:

    >>> import asyncio
    >>> import json
    >>> import os

    >>> import uvicorn
    >>> from pyams_chat_ws.webapp import create_application

    >>> base = os.path.dirname(__file__)

    >>> with open(os.path.join(base, 'etc', 'config.json'), 'r') as config_file:
    ...     config = json.loads(config_file.read())

    >>> loop = asyncio.get_event_loop()
    >>> application = loop.run_until_complete(create_application(config))

    >>> if __name__ == '__main__':
    >>>     uvicorn.run(application, host='0.0.0.0', port=8000)

"""

# pylint: disable=logging-fstring-interpolation

try:
    from elasticapm.contrib.starlette import ElasticAPM, make_apm_client
except ImportError:
    ElasticAPM = make_apm_client = None

from starlette.middleware.authentication import AuthenticationMiddleware

from .app import ChatApp
from .auth import JWTAuthenticationBackend


__docformat__ = 'restructuredtext'


def get_config_keys(config, prefix):
    """Get configuration options matching given prefix"""
    for key, value in config.items():
        if key.startswith(prefix):
            yield key[len(prefix):].upper(), value


async def create_application(config):
    app = await ChatApp.create(config)
    app.add_middleware(AuthenticationMiddleware,
                       backend=JWTAuthenticationBackend(config))
    if ElasticAPM is not None:
        apm_prefix = config.pop('apm_prefix', 'apm_')
        apm_config = dict(get_config_keys(config, apm_prefix))
        service_name = apm_config.get('SERVICE_NAME')
        if service_name:
            apm = make_apm_client(apm_config)
            app.add_middleware(ElasticAPM, client=apm)
    return app



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat_ws.auth module

This module provides JWT authentication backend.
"""

# pylint: disable=logging-fstring-interpolation

import httpx
from starlette.authentication import AuthCredentials, AuthenticationBackend, SimpleUser
from starlette.requests import HTTPConnection

from pyams_chat_ws import LOGGER


__docformat__ = 'restructuredtext'


class JWTUser(SimpleUser):
    """JWT authenticated user"""

    def __init__(self, username: str, access_token: str):
        super().__init__(username)
        self.access_token = access_token


class JWTAuthenticationBackend(AuthenticationBackend):
    """JWT authentication backend"""

    def __init__(self, config: dict):
        super().__init__()
        self.config = config

    async def authenticate(self, request: HTTPConnection):  # pylint: disable=arguments-renamed
        """Request authentication"""
        protocol = request.headers.get('sec-websocket-protocol')
        if not protocol:
            LOGGER.debug('No WebSocket security protocol, authentication refused!')
            return None
        proto, token = protocol.split()
        if ',' in proto:
            proto, _ = proto.split(',', 1)
        if not proto.startswith(self.config.get('ws_protocol', 'accessToken')):
            LOGGER.debug('Unknown security protocol, authentication refused!')
            return None
        authority = self.config.get('jwt_authority')
        LOGGER.debug(f'Checking JWT authority {authority}')
        async with httpx.AsyncClient(verify=self.config.get('ssl_verify', True)) as client:
            result = await client.get(authority, headers={
                'Authorization': f'Bearer {token}',
                'Content-type': 'application/json'
            })
            if result.status_code != httpx.codes.OK:
                LOGGER.debug(f'JWT status code: {result.status_code}')
                return None
            username = result.json().get('sub')
            if username:
                LOGGER.debug(f'JWT username: {username}')
                return AuthCredentials(['authenticated']), \
                    JWTUser(username, token)
        LOGGER.debug("Can't get valid credentials from authentication token!")
        return None



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat_ws.server module

This module defines main chat server application.
"""

# pylint: disable=logging-fstring-interpolation

import asyncio
import json

import aioredis
import async_timeout
import httpx
from starlette.applications import Starlette
from starlette.routing import Route, WebSocketRoute
from starlette.websockets import WebSocket

from pyams_chat_ws import LOGGER
from pyams_chat_ws.chat import WSChatEndpoint
from pyams_chat_ws.monitor import MonitorEndpoint


__docformat__ = 'restructuredtext'


class ChatApp(Starlette):
    """Main chat application"""

    redis = None

    sessions = {}
    sessions_lock = asyncio.Lock()

    @classmethod
    async def create(cls, config):
        """Application factory"""
        app = cls(config)
        app.redis = await aioredis.from_url(config.get('redis_host'),
                                            encoding='utf-8',
                                            decode_responses=True)
        asyncio.create_task(app.start_chat())
        return app

    def __init__(self, config):
        super().__init__(routes=[
            Route(config.get('monitor_endpoint', '/monitor'),
                  MonitorEndpoint),
            WebSocketRoute(config.get('ws_endpoint', "/ws/chat"),
                           WSChatEndpoint)
        ])
        self.config = config

    async def start_chat(self):
        """Chat application starter"""
        psub = self.redis.pubsub()
        await psub.subscribe(self.config.get('channel_name', "chat:main"))
        LOGGER.debug('Redis channel subscription enabled')
        async with psub as p:  # pylint: disable=invalid-name
            while True:
                try:
                    async with async_timeout.timeout(1):
                        msg = await p.get_message(ignore_subscribe_messages=True)
                        if msg is not None:
                            LOGGER.debug(f'Loaded Redis message: {msg!r}')
                            await self.dispatch(msg)
                        await asyncio.sleep(0.01)
                except asyncio.TimeoutError:
                    pass

    async def add_session(self, ws: WebSocket):  # pylint: disable=invalid-name
        """Add session from given websocket"""
        context_url = self.config.get('context_url')
        LOGGER.debug(f'>>> Adding new websocket session...')
        LOGGER.debug(f'  > client: {ws.client}')
        LOGGER.debug(f'  > context URL: {context_url}')
        token = ws.user.access_token
        LOGGER.debug(f'  > token: {token}')
        async with httpx.AsyncClient(verify=self.config.get('ssl_verify', True)) as client:
            result = await client.get(context_url, headers={
                'Authorization': f'Bearer {token}',
                'Content-type': 'application/json'
            })
            LOGGER.debug(f'  > got context: {result}')
            if result.status_code != httpx.codes.OK:
                return None
            context = result.json()
            LOGGER.debug(f'  > context data: {context}')
            async with self.sessions_lock:
                self.sessions[ws.client] = {
                    'ws': ws,
                    'host': ws.headers.get('origin'),
                    'context': context,
                    'principal_id': ws.user.username,
                    'channels': [self.config.get('channel_name', "chat:main")]
                }
                LOGGER.debug(f'>>> users sessions count: {len(self.sessions)}')

    def drop_session(self, ws: WebSocket):  # pylint: disable=invalid-name
        """Drop session from given websocket"""
        LOGGER.debug(f'>>> dropping session for {ws.client}...')
        self.sessions.pop(ws.client, None)

    async def dispatch(self, message):
        """Dispatch received message"""
        LOGGER.debug(">>> dispatching message...")
        if isinstance(message, (str, bytes)):
            try:
                message = json.loads(message)
            except ValueError:
                return
        message = message.get('data')
        if isinstance(message, (str, bytes)):
            try:
                message = json.loads(message)
            except ValueError:
                return
        if message:
            async with self.sessions_lock:
                principals = set(message.pop('target', {}).get('principals', []))
                for session in self.sessions.values():
                    LOGGER.debug(f"  > checking session: {session}")
                    # don't send messages to other hosts
                    if session['host'] != message.get('host'):
                        continue
                    # don't send messages to message emitter
                    if session['principal_id'] == message.get('source', {}).get('id'):
                        continue
                    # filter message targets
                    try:
                        if set(session['context']['principal']['principals']) & principals:
                            LOGGER.debug(f"  > sending message: {message}")
                            await session['ws'].send_json(message)
                    except KeyError:
                        continue
            # update Redis notifications queue
            cache_key = self.config.get('notifications_key')
            if cache_key:
                cache_key = f"{cache_key}::{message.get('host', '--')}"
                cache_length = self.config.get('notifications_length', 50)
                LOGGER.debug(f"   > adding message to Redis queue {cache_key}")
                async with self.redis.pipeline(transaction=True) as pipe:
                    await pipe \
                        .lpush(cache_key, json.dumps(message)) \
                        .ltrim(cache_key, 0, cache_length - 1) \
                        .execute()



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

__docformat__ = 'restructuredtext'

import sys

import pkg_resources
from pkg_resources import DistributionNotFound
from starlette.endpoints import HTTPEndpoint
from starlette.responses import JSONResponse


versions = [
    f'Python/{sys.version_info.major}.{sys.version_info.minor}'
]
for package in ('websockets', 'Starlette'):
    versions.append(f'{package}/{pkg_resources.get_distribution(package).version}')

try:
    distribution = pkg_resources.get_distribution('pyams-chat-ws')
except DistributionNotFound:
    distribution = None

versions.append(f'''PyAMS-chat-ws/{distribution.version 
    if distribution is not None else "development"}''')


class MonitorEndpoint(HTTPEndpoint):
    """Application monitoring endpoint"""

    async def get(self, request):
        """Default monitor endpoint"""
        return JSONResponse({
            'status': 'OK',
            'sessions_count': len(self.scope['app'].sessions),
            'server': ' '.join(versions)
        })



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat_ws.chat module

This module defines a websocket endpoint used for chat.
"""

# pylint: disable=logging-fstring-interpolation

from starlette.authentication import UnauthenticatedUser
from starlette.endpoints import WebSocketEndpoint
from starlette.websockets import WebSocket

from pyams_chat_ws import LOGGER


__docformat__ = 'restructuredtext'


class WSChatEndpoint(WebSocketEndpoint):
    """Main chat endpoint"""

    encoding = 'text'

    async def on_connect(self, ws: WebSocket):  # pylint: disable=arguments-renamed
        """Websocket connection handler"""
        await ws.accept('accessToken')
        if 'authenticated' in ws.auth.scopes:
            app = self.scope.get('app', None)
            if app is not None:
                LOGGER.debug(f'Adding user session for {ws.user.username}...')
                await app.add_session(ws)
        else:
            await ws.send_json({
                'action': 'logout',
                'status': 'FORBIDDEN'
            })

    async def on_disconnect(self, ws: WebSocket, close_code: int):  # pylint: disable=arguments-renamed
        """Websocket disconnection handler"""
        app = self.scope.get('app', None)
        if app is not None:
            if isinstance(ws.user, UnauthenticatedUser):
                LOGGER.debug(f'Dropping session for unauthenticated user...')
            else:
                LOGGER.debug(f'Dropping user session for {ws.user.username}...')
            app.drop_session(ws)

    async def on_receive(self, ws: WebSocket, data):  # pylint: disable=arguments-renamed
        """Websocket message handler"""
        LOGGER.debug(f'Received message: {data}')
        if data == 'PING':
            await ws.send_text('PONG')
            return
        if 'authenticated' in ws.auth.scopes:
            app = self.scope.get('app', None)
            if app is not None:
                await app.dispatch(data)
